/*
 * $RCSfile: ethkern.c,v $
 * 
 * x-kernel v3.2

 * The ETHKERN driver reads packets from the Linux kernel before they
 * are transmitted to the NIT, and write ethernet packets into the IP
 * layer.  This is for use with the ETHDEV driver and the IP forwarder
 * protocols.

 * The driver supports IP address translation so the Linux kernel can
 * keep its real IP address but use a bogus address inside the
 * x-kernel.  The IPAT rom option must is used to enable this feature.
 
 *
 * Copyright (c) 1996  Arizona Board of Regents
 *
 * $Log: ethkern.c,v $
 * Revision 1.5  1997/04/23 19:05:44  mjk
 * Switched to using /dev/htap
 *
 * Revision 1.4  1997/03/10 21:36:41  mjk
 * *** empty log message ***
 *
 * Revision 1.3  1996/11/22 18:42:48  mjk
 * *** empty log message ***
 *
 * Revision 1.1  1996/07/23 23:59:37  mjk
 * Initial revision
 *
 */


#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <net/if.h>
#include <linux/if_ether.h>
#include <linux/ethtap.h>
#include "xkernel.h"
#include "eth.h"
#include "ip.h"
#include "eth_i.h"
#include "machine.h"
#include "inputProcess.h"
#include "romopt.h"
#include "thread.h"
#include "config.h"


/* 
 * Defines 
 */

#define ETHKERN_NUMTHREADS	16	/* # of xk event rx handlers 	*/
					/* max rx packet size 		*/
#define ETHKERN_BUFSIZE		(MAX_ETH_DATA_SZ + 100)

#ifdef XK_DEBUG
/* #define XK_RX_DEBUG */
/* #define XK_TX_DEBUG */
#endif



/* 
 * Typedefs 
 */

typedef struct {		/* Protocol State 			*/
  ETHhost	ethaddr;		/* ethernet address		*/
  IPhost	ip_real;		/* Linux IP address 		*/
  IPhost	ip_bogus;		/* xkernel IP address 		*/
  u_short	ip_delta;		/* ones compliment IPAT delta	*/
  int		ipatp;			/* IP address translation? 	*/
  int		htap;			/* /dev/htap fd 		*/
  pthread_t	tid;			/* recieving thread 		*/
					/* rx thread attributes 	*/
  pthread_attr_t	thread_attributes;
  BufferPool		*buffers;	/* incomming data buffer pool	*/
  int			shutdown;	/* shutdown predicate 		*/
} PState;


/* This is from the IP protocol internal header file.  This is needed
 * so we can hack the IP addresses when using IPAT.
 */
typedef struct {
  	u_char 	vers_hlen;	/* high 4 bits are version, low 4 are hlen */
	u_char  type;
	u_short	dlen;
	u_short ident;
	u_short frag;
	u_char  time;
	u_char  prot;
	u_short checksum;
  	IPhost	source;		/* source address */
  	IPhost	dest;		/* destination address */
} IPheader; 

typedef struct {
    short 	arp_hrd;
    short 	arp_prot;
    char	arp_hlen;
    char  	arp_plen;
    short 	arp_op;
    ETHhost 	arp_sha;
    IPhost	arp_spa;
    ETHhost	arp_tha;
    IPhost	arp_tpa;
} ArpHdr;


/* 
 * Prototypes 
 */

void			ethkern_init(XObj);
static xkern_return_t	ethkern_openenable(XObj, XObj, XObj, Part *);
static xkern_return_t	ethkern_shutdown(XObj);
static xmsg_handle_t	ethkern_push(XObj, Msg *);
static int		ethkern_control(XObj, int, char *, int);

static long		ethkern_msgload(void *, char *, long, void *);
static void		ethkern_internal_demux(InputBuffer *);

static void		*ethkern_rxthread(void *);
static void		ethkern_gethost(XObj);

static xkern_return_t   ethkern_rom_ipat(XObj, char **, int, int, void *);

#ifdef XK_DEBUG
static void		ethkern_dump(XObj, FILE *, const char *,
				     unsigned char *, int);
#endif


/* 
 * Globals 
 */

int			traceethkernp;	/* trace variable 		*/

					/* rom options			*/
static XObjRomOpt	ethkern_opts[] = {
  { "ipat",	4, ethkern_rom_ipat }
};





/* 
 * Public Functions
 */

void
ethkern_init(XObj self) 
{
  PState		*ps;
#ifndef ETHKERN_NOWRITE
  int			mode = ETHTAP_ACTIVE_MODE;
#endif  


  xTraceP0(self, TR_GROSS_EVENTS, "init");
  ps = X_NEW(PState);
  ps->shutdown = 0;
  ps->ipatp    = 0;		/* default is no IP translation 	*/
  self->state = ps;

  				/* Parse ROM file 			*/
  findXObjRomOpts(self, ethkern_opts,
		  (sizeof ethkern_opts) / sizeof(XObjRomOpt), NULL);

#if 0				/* revist this optimization 		*/
  
				/* Compute ones-complement IP header
                                   delta used to correct the header
                                   checksum in ethkern_push(). */
  if ( ps->ipatp ) {
    ps->ip_delta =  ~*((u_short *)&ps->ip_bogus.a) +
		    ~*((u_short *)&ps->ip_bogus.c) +
		    *((u_short *)&ps->ip_real.a) +
		    *((u_short *)&ps->ip_real.c);
    xTraceP1(self, TR_MAJOR_EVENTS, "IPAT delta checksum is %04x\n",
	     ps->ip_delta);
  }
#endif  
  
  if ( (ps->htap = open("/dev/htap", O_RDWR|O_NONBLOCK)) == -1 ) {
    xError("cannot open /dev/htap");
    exit(-1);
  }

#ifdef ETHKERN_NOWRITE
  xTraceP0(self, TR_GROSS_EVENTS, "passive mode");
#else
  if ( ioctl(ps->htap, ETHTAPSETMODE, &mode) ) {
    xError("xkin ioctl ETHTAPSETMODE failed");
    exit(-1);
  }
#endif

  ethkern_gethost(self);	/* read eth0 address 			*/


  ps->buffers = xkBufferPoolInit(ETHKERN_NUMTHREADS, ETHKERN_BUFSIZE,
				 ethkern_internal_demux);


  self->openenable	= ethkern_openenable;
  self->shutdown	= ethkern_shutdown;
  self->control		= ethkern_control;
  self->push		= ethkern_push;
  self->up		= NULL;
  

} /* ethkern_init */



/* 
 * Public Protocol Methods 
 */


static xkern_return_t
ethkern_openenable(XObj self, XObj hlpRcv, XObj hlpType, Part *p)
{
  struct sched_param	sched;
  PState		*ps = self->state;

  if ( self->up ) {
    xError("ethkern_openenable called multiple times!");
    return XK_FAILURE;
  }
  
  self->up = hlpRcv;

				/* Start rxer thread 			*/
  pthread_attr_init(&ps->thread_attributes);
  pthread_attr_getschedparam(&ps->thread_attributes, &sched);
  sched.sched_priority = THREAD_STDPRIO - 10; /* NOT POSIX		*/
  pthread_attr_setschedparam(&ps->thread_attributes, &sched);
  pthread_create(&ps->tid, &ps->thread_attributes, ethkern_rxthread, self);
  
  return XK_SUCCESS;
} /* ethpth_openable */


static xkern_return_t
ethkern_shutdown(XObj self) 
{
  PState		*ps = self->state;

  xTraceP0(self, TR_MAJOR_EVENTS, "shutdown");
  ps->shutdown = 1;		/* shutdown driver threads 		*/
  close(ps->htap);

  return XK_SUCCESS;
} /* ethkern_shutdown */


static int
ethkern_control(XObj self, int opcode, char *buf, int buflen)
{
  PState	*ps = (PState *)self->state;

  switch ( opcode ) {

    case GETMAXPACKET:
      xTraceP0(self, TR_MAJOR_EVENTS, "getmaxpacket");
      checkLen(buflen, sizeof(int));
      *(int *)buf = MAX_ETH_DATA_SZ;
      return sizeof(int);
      
    case GETMYHOST:
      xTraceP0(self, TR_MAJOR_EVENTS, "getmyhost");
      checkLen(buflen, sizeof(ETHhost));
      memcpy(buf, &ps->ethaddr, sizeof(ETHhost));
      return sizeof(ETHhost);

    case ETH_SETPROMISCUOUS:
      xTraceP0(self, TR_MAJOR_EVENTS, "setpromiscuous");
      return 0;

    default:
      return -1;
  }
} /* ethpkt_control */


static xmsg_handle_t
ethkern_push(XObj self, Msg *packet)
{
  PState		*ps 	= (PState *)self->state;
  ETHhdr		*header	= msgGetAttr(packet, 0);
  unsigned char		buf[ETHKERN_BUFSIZE];
  unsigned char		*numpkts = buf;
  unsigned short	*pktlen  = (unsigned short *)&buf[1];
  unsigned char		*data    = (unsigned char *)&buf[3];
  int			sentlen;

  xTraceP0(self, TR_EVENTS, "push");
 
  *numpkts = 1;			/* write one packet at a time 		*/

  if ( !msgGetAttr(packet, 0) ) {
    xTraceP0(self, TR_ERRORS, "push -- no header attached");
    return XMSG_ERR_HANDLE;
  }

  if ( (*pktlen = msgLen(packet)) > ETHKERN_BUFSIZE ) {
    xTraceP1(self, TR_ERRORS, "push -- msgLen(%d) too big", msgLen(packet));
    return XMSG_ERR_HANDLE;
  }

				/* Create a flat buffer with the
                                   complete ethernet packet */
  memcpy(data, header, sizeof(ETHhdr));

  xTraceP4(self, TR_MAJOR_EVENTS, "push: %04x %s > %s %d bytes",
	   ntohs(header->type),
	   ethHostStr(&header->src), ethHostStr(&header->dst),
	   msgLen(packet));

  msg2buf(packet, &data[sizeof(ETHhdr)]);
  *pktlen += sizeof(ETHhdr);	/* add header to length 		*/
#ifdef ETHKERN_PADOUT  
  if ( *pktlen < ETH_ZLEN )	/* pad to minimum length (w/o FCS)	*/
    *pktlen = ETH_ZLEN;
#endif  

				/* Translate IP dest and fixed
                                   checksum */
  if ( ps->ipatp ) {
    ETHhdr	*eth = (ETHhdr *)data;

    if ( eth->type == htons(0x0800) ) {
      IPheader	*p;

      p = (IPheader *)(data + sizeof(ETHhdr));
      memcpy(&p->dest, &ps->ip_real, sizeof(IPhost));

      p->checksum = 0x0000;
      p->checksum = ~ocsum((u_short *)p, sizeof(IPheader) / 2);
    }
    
    else if ( eth->type == htons(0x0806) ) {
      ArpHdr	*p;

      p = (ArpHdr *)(data + sizeof(ETHhdr));
      memcpy(&p->arp_tpa, &ps->ip_real, sizeof(IPhost));
    }
    
  }

#ifdef XK_TX_DEBUG
  xIfTraceP(self, TR_EVENTS) {
    ethkern_dump(self, stdout, "TX", numpkts,
		 *pktlen + sizeof(char) + sizeof(short));
  }
#endif  

  sentlen = write(ps->htap, numpkts, *pktlen + sizeof(char) + sizeof(short));

  if ( sentlen == -1 ) {
    xTraceP0(self, TR_ERRORS, "push: write failed");
    return XMSG_ERR_HANDLE;
  }

  xTraceP1(self, TR_MORE_EVENTS, "msg sent %d bytes", sentlen);

  return XMSG_NULL_HANDLE;
} /* ethkern_push */



/* 
 * Private Protocol Methods 
 */

static void
ethkern_internal_demux(InputBuffer *buffer)
{
  ETHhdr	header;
  Msg		*msg	= &buffer->msg;
  XObj		self	= buffer->self;

  if ( !msgPop(msg, ethkern_msgload, &header, sizeof(ETHhdr), NULL) ) {
    xTraceP0(self, TR_SOFT_ERRORS,
	     "demux: incoming message too small!");
    msgDestroy(msg);
    return;
  }
  if ( !self->up ) {
    xTraceP0(self, TR_ERRORS, "demux: no upper protocol!");
    msgDestroy(msg);
    return;
  }

				/* dump packet header			*/
  xTraceP4(self, TR_MAJOR_EVENTS, "demux: %04x %s > %s %d bytes",
	   ntohs(header.type),
	   ethHostStr(&header.src), ethHostStr(&header.dst),
	   msgLen(msg));

  msgSetAttr(msg, 0, &header, sizeof(ETHhdr));

  
  xDemux(self, msg);		/* send to eth protocol 		*/
} /* ethkern_internal_demux */


static long
ethkern_msgload(void *header, char *buf, long buflen, void *arg )
{
  xAssert(buflen == sizeof(ETHhdr));
  memcpy(header, buf, sizeof(ETHhdr));
  return sizeof(ETHhdr);
} /* ethkern_msgload */



/* 
 * Private ROM file methods 
 */

static xkern_return_t 
ethkern_rom_ipat(XObj self, char **argv, int argc, int line, void *arg)
{
  PState	 *ps = (PState *)self->state;

  if ( argc == 4 ) {
    if ( str2ipHost(&ps->ip_real, argv[2]) == XK_SUCCESS &&
	 str2ipHost(&ps->ip_bogus, argv[3]) == XK_SUCCESS )
    {
      ps->ipatp = 1;		/* address translation enabled 		*/
      xTraceP1(self, TR_MAJOR_EVENTS, "linux host is %s",
	       ipHostStr(&ps->ip_real));
      xTraceP1(self, TR_MAJOR_EVENTS, "xkernel host is %s",
	       ipHostStr(&ps->ip_bogus));
      return XK_SUCCESS;
    }
  }
  return XK_FAILURE;
} /* ethkern_rom_ipat */



/* 
 * Private non-xKernel driver method.
 */

				/* The following routines are called
                                   outside of xKernel control so
                                   careful use of xk_master_lock() and
                                   xk_master_unlock() is required. */

static void *
ethkern_rxthread(void *arg)
{
  XObj			self    = (XObj)arg;
  PState		*ps	= (PState *)self->state;

  xk_master_lock();		/* Lock xKernel				*/

  while ( ! ps->shutdown ) {
    unsigned char	buf[0x4000];
    unsigned char	*p 	= buf;
    unsigned char	pkts;
    unsigned short	len;
    InputBuffer		*buffer;
    int			i;
    
    if ( read(ps->htap, buf, sizeof buf) < 0 ) {
      xError("ethkern ERROR: Cannot read from kernel");
      continue;
    }
    
    pkts = *p++;
    if ( pkts ) {
      xTraceP1(self, TR_EVENTS, "read %d packets", pkts);
      for (i=0; i<pkts; i++) {
	len = *(unsigned short *)p;
	p  +=  sizeof len;
	if ( !(buffer = xkBufferPoolNextBlock(ps->buffers) ) ) {
	  xError("ethkern ERROR: Can't get buffer, dropping incoming packet");
	  xIfTraceP(self, TR_SOFT_ERRORS) {
	    xkBufferPoolDump(ps->buffers);
	  }
	}
	else {
	  memcpy(buffer->data, p, len);
	  msgTruncate(&buffer->msg, len);
	  buffer->self = self;
	  xTraceP1(self, TR_FULL_TRACE,
		   "sem count (pre signal) == %d", buffer->sem.count);
#ifdef XK_RX_DEBUG
	  xIfTraceP(self, TR_EVENTS) {
	    ethkern_dump(self, stdout, "RX", p, len);
	  }
#endif  
	  semSignal(&buffer->sem);
	}
	p += len;
      }
    }
    Yield();			/* pass control to new packets 		*/
  }
  
  xk_master_unlock();		/* UnLock xKernel 			*/
      
  return NULL;
} /* ethkern_rxthread */


static void
ethkern_gethost(XObj self) 
{
  PState		*ps = self->state;
  int			fd;
  struct ifreq		ifr;
  
  if ( (fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ) {
    xError("cannot open socket");
    exit(1);
  }

  strcpy(ifr.ifr_name, "eth0");
  if ( ioctl(fd, SIOCGIFHWADDR, &ifr) == -1 ) {
    xTraceP0(self, TR_ERRORS, "SIOCGIFHWADDR ioctl failed");
    exit(1);
  }

  memcpy(&ps->ethaddr, &ifr.ifr_hwaddr.sa_data, sizeof(ETHhost));
  xTraceP1(self, TR_GROSS_EVENTS, "interface address %s",
	   ethHostStr(&ps->ethaddr));

} /* ethkern_gethost */


#ifdef XK_DEBUG
static void
ethkern_dump(XObj self, FILE *fout, const char *label,
	     unsigned char *pkt, int len)
{
  PState	*ps = (PState *)self->state;
  int		i, j;

  xAssert(fout);
  xAssert(pkt);
  xAssert(label);
  xAssert(len >= 0);

	  /* Dumping is in the debug hex/ascii style and is setup for
	     80 columns.  This puts 16 bytes per line. */
 
/*   len = min(len, ps->dumplen); */
	    
  for (i=0; i<len; i+=16) {
    char	buffer[80];
    char	*p = &buffer[0];
    char	*q = &buffer[52];
    
    memset(buffer, (int)' ', sizeof buffer);
    
    for (j=0; j<16 && (j+i) < len; j++) {
      unsigned char	byte = pkt[i+j];

      if ( j == 8 )
	p += sprintf(p, "- ");
      
      p += sprintf(p, "%02x ", byte);
      
      if ( isprint(byte) )
	*q++ = byte;
      else
	*q++ = '.';
    }

    *p = ' ';
    *q = '\0';
    xTraceP2(self, TR_GROSS_EVENTS, "%s: %s", label, buffer);
  }
} /* ethkern_dump */
#endif

/* ethkern.c ends here */
 


