
/* Open issues, some marked in the code with a $$.
   Provide for releasing the xk master lock while computing.
   Byte order note:  Only Big Endian testing so far.
   xkrandom:  Drop it for now, but leave hooks in.
   No provision for defending against the "man in the middle" attack.
     This requires adding signatures.
   Perhaps dhkxClose should invalidate & erase the keys in the KMGR sessions.
   In the interest of speed, dhkxKeyExchangeReply has a timing glitch: it
     calculates & sends the reply message before calculating & setting the key.
     Another message might arrive before we've set the key; it will be dropped.
 */

/* DHECKX misc design notes:

   RVRs mods to event code require "returning" semaphores.  I'm ignoring this.

   I decided not to do retries for the present.  If the offer message is lost,
   the code will time out after ten seconds and the Open will return failure.
   If I were to do retries, I should remember recent offers, in case an old
   one is accepted.

   If an Open is attempted on a session that is being set up, it fails.
   Arguably it should be queued.

   OpenDone at higher level can't send any messages until KX has processed
   first popped message with KX offer.

   Certain kinds of surprize Close are impossible because noone has a
   pointer to the current session.

   Future versions might provide for use of primes > 512 bits.
*/


/* #define TIME */  /* Turns on timing code.  Prints times to do key exchange. */

#include "xkernel.h"
#include "dheckx.h"
#include "dheckx_i.h"
#include "eth.h"		/* So ARP can go on top of DHECKX */
#include "ip.h"                 /* For IP pseudoheader fixup. */
#include "km.h"

extern int tracedheckxp; int tracedheckxp;


#ifdef __STDC__

static XObj		dheckxCreateSessn( XObj, XObj, XObj, XObj);
static void		dheckxSessnInit( XObj );
static int		dheckxControlProtl( XObj, int, char *, int );
static xkern_return_t	dheckxDemux( XObj, XObj, Msg * );
static XObj		dheckxOpen( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxOpenEnable( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxOpenDisable( XObj, XObj, XObj, Part * );
static xkern_return_t	dheckxOpenDone( XObj, XObj, XObj, XObj);
static xkern_return_t	dheckxClose( XObj );
static xmsg_handle_t	dheckxPush( XObj, Msg * );
static xkern_return_t	dheckxPop( XObj, XObj, Msg *, void * );
static int		dheckxControlSessn( XObj, int, char *, int );
static void		dheckxKeyExchangeInitiate(XObj);
static xkern_return_t	dheckxKeyExchangeReply(XObj, Msg *);
static xkern_return_t	dheckxKeyExchangeDone(XObj, Msg *);
static void		dheckxOpenWakeup(Event, void *);
/* int			xkrandom(unsigned char *, long); */
static long		dheckxcopy(void *, char *, long, void *);

#else

static XObj		dheckxCreateSessn();
static void		dheckxSessnInit();
static int		dheckxControlProtl();
static xkern_return_t	dheckxDemux();
static XObj		dheckxOpen();
static xkern_return_t	dheckxOpenEnable();
static xkern_return_t	dheckxOpenDisable();
static xkern_return_t	dheckxOpenDone();
static xkern_return_t	dheckxClose();
static xmsg_handle_t	dheckxPush();
static xkern_return_t	dheckxPop();
static int		dheckxControlSessn();
static void		dheckxKeyExchangeInitiate();
static xkern_return_t	dheckxKeyExchangeReply();
static xkern_return_t	dheckxKeyExchangeDone();
static void		dheckxOpenWakeup();
/* int			xkrandom(); */
static long		dheckxcopy();

#endif /* __STDC__ */

#ifdef TIME
static void subtime(t1, t2, t3)    XTime *t1, *t2, *t3;
{   t3->sec = t2->sec - t1->sec;
    t3->usec = t2->usec - t1->usec;
    if (t3->usec < 0) { t3->usec += 1000000; t3->sec -= 1; } }
#endif

/* Initiate a key exchange.
   Called from Open, with the session (in state DHECKXINIT) as an argument.
   Composes a key exchange initiation message, and ships it.
   Sets session state to DHECKXINITWAIT to indicate we are waiting for a reply
   message to complete the exchange. */
static void dheckxKeyExchangeInitiate(sessn) XObj sessn;
{ SState *ss;
  unsigned long msgbuf[1+1+(2*DEGW)];
  unsigned long expo[DEGW] = FIVE( 0, 0, 0, 36264332u, 1256789 );
  Msg msg;

  xTrace1(dheckxp, TR_MAJOR_EVENTS, "dheckxKeyExchangeInitiate entered: session = %x", sessn);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dheckxp, TR_MORE_EVENTS, "dheckxKeyExchangeInitiate: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHECKXINIT);

#ifdef TIME
  xGetTime(&ss->starttime);
#endif

  /* Get 512 random bits. */
  /* $$ Temporarily disconnect xkrandom, till I make it more robust. */
  /* if (xkrandom((unsigned char *)xbuf, 16 * sizeof(long))) printf("xkrandom failed"); */

  memcpy(ss->k,expo,DEGW * sizeof(unsigned long));  

  bzero((char *)expo,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxp, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitiate:\nexpo =");
   for (i=0;i<DEGW;i++) printf("%x:",ss->k[i]); printf("\n"); };

  DheckxKnown(ss->k,ss->kpx,ss->kpy);

  xIfTrace(dheckxp, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitiate: k*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->kpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->kpy[i]); printf("\n"); };

  /* build message with DHECKXMAGIC, state DHECKXINIT, gx padded to 512 bits. */
  msgbuf[0] = DHECKXMAGIC; msgbuf[1] = ss->state;

  memcpy(&msgbuf[2],ss->kpx,DEGW * sizeof(unsigned long));
  memcpy(&msgbuf[2+DEGW],ss->kpy,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxp, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n"); };

  /* ship it, destroy message */
  msgConstructBuffer(&msg,(char *)msgbuf, (1+1+2*DEGW) * sizeof(unsigned long));
  xPush(xGetDown(sessn,TRANSPORT_PROTOCOL),&msg);
  msgDestroy(&msg);

  ss->state = DHECKXINITWAIT;  /* set state to DHECKXINITWAIT */
  xTrace1(dheckxp, TR_EVENTS, "dheckxKeyExchangeInitiate exit: state = %d", ss->state);
}


/* Receive a key exchange initiation message, and reply.
   Called from Demux, with the session (in state DHECKXRESPONDER) and setup message
   as arguments.
   Computes the session key based on the setup message and our random data;
   compose and send an appropriate reply to the initiator, for him to
   complete his side of the key setup.
   Enter the computed key in the KM sessions.
   Set session state to DHECKXTRANSPARENT, putting us in transparent mode. */
static xkern_return_t dheckxKeyExchangeReply(sessn,msg) XObj sessn; Msg *msg;
{ SState *ss;
  PState *ps = sessn->myprotl->state;
  unsigned long msgbuf[1+1+(2*DEGW)];
  unsigned long expo[DEGW] = FIVE( 0, 0, 0, 36264832u, 1234789 );
  unsigned char key[8];
  int j,t,tmp; Msg rmsg;

  xTrace2(dheckxp, TR_MAJOR_EVENTS,
	  "dheckxKeyExchangeReply entered: session = %x, msg = %x", sessn, msg);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dheckxp, TR_MORE_EVENTS, "dheckxKeyExchangeReply: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHECKXRESPONDER);

#ifdef TIME
  xGetTime(&ss->starttime);
#endif

  /* Check message length; unpack it. */
  xTrace1(dheckxp, TR_DETAILED, "dheckxKeyExchangeReply: msgLen = %d", msgLen(msg));
  if ((tmp = msgLen(msg)) != sizeof(unsigned long) * (1+1+(2*DEGW)))
  { xTrace4(dheckxp, TR_GROSS_EVENTS,
	"dheckxKeyExchangeReply: Session %x ss %x: Offer message %x length %d is wrong",
	    sessn, ss, msg, tmp);
    return XK_FAILURE; };
  msgPop(msg, dheckxcopy, (char *) msgbuf, (1+1+(2*DEGW))*sizeof(unsigned long), 0);

  xIfTrace(dheckxp, TR_DETAILED)
  {  int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n");};

  /* Check msg starts with DHECKXMAGIC, and state of sender is DHECKXINIT. */
  if (!(msgbuf[0] == DHECKXMAGIC && msgbuf[1] == DHECKXINIT))
  { xTrace5(dheckxp, TR_GROSS_EVENTS,
	    "dheckxKeyExchangeReply: Session %x ss %x: Offer message %x."
	    " Magic %d or SendState %d is wrong",
	    sessn, ss, msg, msgbuf[0], msgbuf[1]);
    return XK_FAILURE; };

  memcpy(ss->kpx,&msgbuf[2],DEGW * sizeof(unsigned long));
  memcpy(ss->kpy,&msgbuf[2+DEGW],DEGW * sizeof(unsigned long));


  /* Call xkrandom to get y, 512 bits. */
  /* $$ xkrandom is temporarily turned off. */
  /* if (xkrandom((unsigned char *)ybuf, 16 * sizeof(long))) printf("xkrandom failed"); */

  memcpy(ss->k,expo,DEGW * sizeof(unsigned long));  

  bzero((char *)expo,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxp, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeInitiate:\nexpo =");
   for (i=0;i<DEGW;i++) printf("%x:",ss->k[i]); printf("\n"); };

  DheckxKnown(ss->k,ss->jpx,ss->jpy);


  xIfTrace(dheckxp, TR_DETAILED)
  { int i; printf("dheckxKeyExchangeInitatiate: j*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpy[i]); printf("\n"); };

  ss->state = DHECKXREPLYING;  /* Advance session state to indicate we are replying. */

  /* reply message: DHECKXMAGIC, state = DHECKXREPLYING, gx & gy each padded to 512 */
  /* DHECKXMAGIC & gx are already present in msgbuf */

  msgbuf[1] = ss->state;

  memcpy(&msgbuf[2],&ss->jpx,DEGW * sizeof(unsigned long));
  memcpy(&msgbuf[2+DEGW],&ss->jpy,DEGW * sizeof(unsigned long));

  xIfTrace(dheckxp, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n");};

  /* Ship the reply.
  $$ Timing glitch: xPush releases the master lock.
     I could get another msg before I've finished computing and initialized
     the encryption keys.
     As long as the encryption protocol is above me in the graph, I can just
     throw away messages that arrive before I've finished computing the key.
     Presumably the correct thing to do is queue them. */

  msgConstructBuffer(&rmsg,(char *)msgbuf,(1+1+2*DEGW) * sizeof(unsigned long));
  xPush(xGetDown(sessn,TRANSPORT_PROTOCOL),&rmsg);
  msgDestroy(&rmsg);  /* destroy message */

#ifdef TIME
  xGetTime(&ss->endtime);
#endif

  Dheckx(ss->kpx,ss->kpy,ss->k,ss->jpx,ss->jpy);

  memcpy(key,ss->jpx,4);
  memcpy(&key[4],ss->jpy,4);

  xIfTrace(dheckxp, TR_DETAILED)
  { int i; printf("dheckxKeyFinalKey: k*j*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpy[i]); printf("\n"); };

  /* fix parity on key - each byte should be odd */
  for (j=0;j<8;j++) { t = key[j]; t ^= t>>4; t ^= t>>2; t ^= t>>1; key[j] ^= (~t)&1; }
  xIfTrace(dheckxp, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeReply: odd parity key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }

  /* Enter the key in two KM sessions, one for each host. */
  if (ps->flags & ALL_PARTS)
  { xTrace1(dheckxp, TR_DETAILED, "dheckxKeyExchangeReply: key session is %x",
	    xGetDown(sessn,KEY_MGR_SESS_ALL)); }
  else
  { xTrace2(dheckxp, TR_DETAILED, "dheckxKeyExchangeReply: key sessions are %x and %x",
	  xGetDown(sessn,KEY_MGR_SESS_IN), xGetDown(sessn,KEY_MGR_SESS_OUT)); }
  if (!(ps->flags & ALL_PARTS)) 
    xControl(xGetDown(sessn,KEY_MGR_SESS_IN), KM_SET, (char *)key, 8);
  xControl(xGetDown(sessn,KEY_MGR_SESS_OUT), KM_SET, (char *)key, 8); /* OUT==ALL */
  bzero((char *)key,8); t=0;  /* Zero out key array on stack. */
  /* At this point, I don't really need to remember anything: the key has been
     handed off to the key manager, and all my intermediate values can be trashed. */

  ss->state = DHECKXTRANSPARENT;  /* go transparent */

#ifdef TIME
  xGetTime(&ss->endtime2);
  subtime(&ss->starttime,&ss->endtime,&ss->deltatime);
  subtime(&ss->starttime,&ss->endtime2,&ss->deltatime2);
  printf("\nResponder replied after %d.%06d seconds, and finished key calculation after %d.%06d seconds.\n",
     ss->deltatime.sec, ss->deltatime.usec, ss->deltatime2.sec, ss->deltatime2.usec);
#endif

  xTrace1(dheckxp, TR_EVENTS, "dheckxKeyExchangeReply exit: state = %d", ss->state);
  return XK_SUCCESS;
}



/* Handle the reply to the key exchange initiation message.
   Called from Demux, with the session (in state DHECKXINITWAIT) and reply message
   as arguments.
   Check that reply message matches our key initiation message.
   Set session state to DHECKXCOMPUTING.
   Computes the session key based on our initiation data and the
   other information in the reply message.
   Enter the computed key in the KM sessions.
   Set session state to DHECKXTRANSPARENT, putting us in transparent mode. */
static xkern_return_t dheckxKeyExchangeDone(sessn,msg) XObj sessn; Msg *msg;
{ SState *ss;
  PState *ps = sessn->myprotl->state;
  unsigned long msgbuf[1+1+(2*DEGW)];
  unsigned char key[8];
  int j,t,tmp;

  xTrace2(dheckxp, TR_MAJOR_EVENTS,
	  "dheckxKeyExchangeDone entered: session = %x, msg = %x", sessn, msg);
  xAssert(xIsSession(sessn));
  ss = sessn->state;
  xTrace2(dheckxp, TR_MORE_EVENTS, "dheckxKeyExchangeDone: ss = %x, state = %d",
	  ss, ss->state);
  xAssert(ss->state == DHECKXINITWAIT);

  /* Check message length; unpack it. */
  xTrace1(dheckxp, TR_DETAILED, "dheckxKeyExchangeDone: msgLen = %d", msgLen(msg));
  if ((tmp = msgLen(msg)) != (1+1+2*DEGW)*sizeof(unsigned long))
  { xTrace4(dheckxp, TR_GROSS_EVENTS,
	"dheckxKeyExchangeDone: Session %x ss %x: Reply message %x length %d is wrong",
	    sessn, ss, msg, tmp);
    return XK_FAILURE; };
  msgPop(msg, dheckxcopy, (char *) msgbuf, (1+1+2*DEGW)*sizeof(unsigned long), 0);

  xIfTrace(dheckxp, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeInitiate: msg is ");
   for (i=0;i<1+1+(2*DEGW);i++)
     printf("%x:",msgbuf[i]); printf("\n");};

  /* Check msg starts with DHECKXMAGIC, and state of sender is DHECKXREPLYING. */
  if (!(msgbuf[0] == DHECKXMAGIC && msgbuf[1] == DHECKXREPLYING))
  { xTrace5(dheckxp, TR_GROSS_EVENTS,
	    "dheckxKeyExchangeDone: Session %x ss %x: Reply message %x."
	    " Magic %d or SendState %d is wrong",
	    sessn, ss, msg, msgbuf[0], msgbuf[1]);
    return XK_FAILURE; };

  memcpy(ss->kpx,&msgbuf[2],DEGW * sizeof(unsigned long));
  memcpy(ss->kpy,&msgbuf[2+DEGW],DEGW * sizeof(unsigned long));


 
  ss->state = DHECKXCOMPUTING;  /* Now committed to the Open. */

  Dheckx(ss->kpx,ss->kpy,ss->k,ss->jpx,ss->jpy);

  memcpy(key,ss->jpx,4);
  memcpy(&key[4],ss->jpy,4);

  xIfTrace(dheckxp, TR_DETAILED)
  { int i; printf("dheckxKeyFinalKey: k*j*p = "); 
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpx[i]);
  for (i=0;i<DEGW;i++) printf("%x:",ss->jpy[i]); printf("\n"); };

  /* set key from gxy */
  xIfTrace(dheckxp, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeDone: key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }

  /* fix parity on key - each byte should be odd */
  for (j=0;j<8;j++) { t = key[j]; t ^= t>>4; t ^= t>>2; t ^= t>>1; key[j] ^= (~t)&1; }
  xIfTrace(dheckxp, TR_DETAILED)
  {int i; printf("dheckxKeyExchangeDone: odd parity key = ");
   for (i=0;i<8;i++) printf(" %02x", key[i]); printf("\n"); }

  /* Enter the key in two KM sessions, one for each host. */
  if (ps->flags & ALL_PARTS)
  { xTrace1(dheckxp, TR_DETAILED, "dheckxKeyExchangeDone: key session is %x",
	    xGetDown(sessn,KEY_MGR_SESS_ALL)); }
  else
  { xTrace2(dheckxp, TR_DETAILED, "dheckxKeyExchangeDone: key sessions are %x and %x",
	  xGetDown(sessn,KEY_MGR_SESS_IN), xGetDown(sessn,KEY_MGR_SESS_OUT)); }
  if (!(ps->flags & ALL_PARTS))
    xControl(xGetDown(sessn,KEY_MGR_SESS_IN), KM_SET, (char *)key, 8);
  xControl(xGetDown(sessn,KEY_MGR_SESS_OUT), KM_SET, (char *)key, 8); /* OUT==ALL */
  bzero((char *)key,8); t=0;  /* Zero out key array on stack. */

  /* There's a timing problem here if the encryption module is below us in the
   graph: there might be an immediate 2nd message awaiting us.  This is unlikely,
   but might happen as a "one-shot" RPC reply.  [If our offer message were bundled
   with an RPC, or an implied RPC.  We don't presently do this.]  In fact, the
   2nd message might even have raced ahead of the KX reply, and been discarded.
   This is all theory right now; but it's a good idea to keep the encryption module
   is above us in the graph.
  */

  /* At this point, I don't really need to remember anything: the key has been
     handed off to the key manager, and all my intermediate values can be trashed. */

  ss->state = DHECKXTRANSPARENT;  /* go transparent */

#ifdef TIME
  xGetTime(&ss->endtime);
  subtime(&ss->starttime,&ss->endtime,&ss->deltatime);
  printf("\nKey exchange completed after %d.%06d seconds.\n",
	 ss->deltatime.sec, ss->deltatime.usec);
#endif

  xTrace1(dheckxp, TR_EVENTS, "dheckxKeyExchangeDone exit: state = %d", ss->state);
  return XK_SUCCESS;
}


static xkern_return_t romfinddelay(XObj, char **, int, int, void *);
static xkern_return_t romfindpartnum(XObj, char **, int, int, void *);
static xkern_return_t romfindhlptype(XObj, char **, int, int, void *);

static XObjRomOpt rmopts[3] = {{ "delay", 1, romfinddelay },
			       { "numparts", 1, romfindpartnum},
			       { "ownhlptype", 0, romfindhlptype}
			       };

static xkern_return_t
romfinddelay(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->delay = atoi(ptr[2]);
  return XK_SUCCESS;
}

static xkern_return_t
romfindhlptype(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  ((PState *)(self->state))->ownhlptype = 1;
  return XK_SUCCESS;
}

static xkern_return_t
romfindpartnum(self, ptr, fields, line, arg)
XObj self;
char **ptr;
int  fields, line;
void *arg;
{
  if (atoi(ptr[2]) > 1)
    ((PState *)(self->state))->flags |= ALL_PARTS;
  else
    ((PState *)(self->state))->flags &= ~ALL_PARTS;
  return XK_SUCCESS;
}

void
dheckx_init( self )
     XObj self;
{
    PState	*ps;
    XObj	llp;

    xTrace1(dheckxp, TR_ERRORS, "dheckx_init entered: protocol = %x", self);
    xAssert(xIsProtocol(self));
    

    if ( ! xIsProtocol(llp = xGetDown(self, TRANSPORT_PROTOCOL)) )
    { xError("dheckx_init: dheckx cannot get lower protocol object"); return; }

    ps = X_NEW(PState);
    self->state = (VOID *)ps;
    ps->actMap = mapCreate(DHECKX_ACTIVE_MAP_SZ, sizeof(ActiveId));
    ps->pasMap = mapCreate(DHECKX_PASSIVE_MAP_SZ, sizeof(PassiveId));
    ps->delay = DHECKX_TIMEOUT;
    ps->flags = 0;
    ps->ownhlptype = 0;


    self->control = dheckxControlProtl;
    self->open = dheckxOpen;
    self->openenable = dheckxOpenEnable;
    self->opendisable = dheckxOpenDisable;
    self->opendone = dheckxOpenDone;
    self->close = dheckxClose;  
    self->demux = dheckxDemux;
    findXObjRomOpts(self, rmopts, 3, 0);
    xTrace0(dheckxp, TR_EVENTS, "dheckx_init exiting");
}

/* Wake up dheckxOpen: Either the session timed out, or the reply arrived.
   evnt is unused, so that Demux can call the same wakeup function. */
static void
dheckxOpenWakeup(evnt, session) Event evnt; void *session;
{ SState	*ss;
  xTrace2(dheckxp,TR_EVENTS,"dheckxOpenWakeup called; evnt = %x, session = %x", evnt, session);
  xAssert(xIsSession((XObj)session));
  ss = ((XObj)session)->state;
  evCancel(ss->event);  /* cancel timeout event - might be our thread, doesn't matter */
  ss->event = (Event) 0;
  semSignal(&ss->semaphore);
 }

static XObj
dheckxOpen( self, hlpRcv, hlpType, part )
    XObj	self, hlpRcv, hlpType;
    Part 	*part;
{
    XObj 	dheckxSessn;
    XObj	lls;
    PState	*pstate;
    SState	*ss;

    xTrace4(dheckxp,TR_EVENTS,"dheckxOpen entered with args %x %x %x %x",
	    self,hlpRcv,hlpType,part);

    pstate = (PState *) self->state;
    
    if ( part == 0 || partLen(part) < 1 )
  { xTrace0(dheckxp, TR_SOFT_ERRORS, "dheckxOpen -- bad participants");   return ERR_XOBJ; }

    xTrace4(dheckxp,TR_EVENTS,"dheckxOpen calling xOpen with args %x %x %x %x",
	    self,hlpType,xGetDown(self,TRANSPORT_PROTOCOL),part);

    lls = xOpen(self, pstate->ownhlptype?self:hlpType,
		xGetDown(self, TRANSPORT_PROTOCOL), part);
    if ( lls == ERR_XOBJ )
    { xTrace0(dheckxp, TR_ERRORS, "dheckxOpen: could not open lls"); return ERR_XOBJ; }

    if ( mapResolve(pstate->actMap, &lls, &dheckxSessn) == XK_SUCCESS )
    { xTrace0(dheckxp, TR_MAJOR_EVENTS, "found an existing one"); xClose(lls); }
    else
    { xTrace0(dheckxp, TR_MAJOR_EVENTS, "creating a new one");
      if ( (dheckxSessn = dheckxCreateSessn(self, hlpRcv, hlpType, lls)) == ERR_XOBJ )
      { xClose(lls);   return ERR_XOBJ; } }

    ss = (SState *)dheckxSessn->state;
    if (ss->state != DHECKXNEW && ss->state != DHECKXTRANSPARENT)
    { xTrace1(dheckxp, TR_ERRORS,
	      "dheckxOpen: tried to do second Open while first in-progress, state = %d",
	      ss->state);
      return ERR_XOBJ; };

    /* If we got a fresh session, start the key exchange protocol.
       Send off an offer message. */
    if (ss->state == DHECKXNEW)
    { ss->state = DHECKXINIT;
      xDuplicate(dheckxSessn); /* Bump the reference count on the session, since
	     it will have message traffic prior to returning from dheckxOpen. */
      dheckxKeyExchangeInitiate(dheckxSessn);
  loop:
      xTrace2(dheckxp, TR_EVENTS, "dheckxOpen: sleeping for %d seconds; ss = %x",
	      pstate->delay, ss);
      /* schedule time out after DHECKX_TIMEOUT seconds */
      ss->event = evSchedule(dheckxOpenWakeup,dheckxSessn,pstate->delay*1000000);
      semInit(&ss->semaphore,0);
      semWait(&ss->semaphore);  /* the pause that refreshes */
      xTrace2(dheckxp, TR_EVENTS, "dheckxOpen: awakened; ss = %x, state = %d",ss,ss->state);
      /* When we get here, either timed out, or got reply. */
      if (ss->state == DHECKXCOMPUTING)
      { xTrace1(dheckxp,TR_EVENTS,
		"dheckxOpen: reply received; still computing key.  ss = %x",ss);
	goto loop; }; /* We are willing to wait forever if computing. */
      if (ss->state == DHECKXINITWAIT)  /* Timed out.  Clean up session & fail. */
      { xTrace1(dheckxp,TR_ERRORS,"dheckxOpen timed out, returning failure. ss = %x",ss);
	dheckxSessn->rcnt--;  dheckxClose(dheckxSessn);  return ERR_XOBJ; };
      if (ss->state != DHECKXTRANSPARENT)  /* Bogus state: bomb. */
      { printf("\ndheckxOpen:  Session state %x has bad state value %d\n", ss, ss->state);
	Kabort("Aborting from dheckxOpen."); };
      /* OK!  Drop through. */
      dheckxSessn->rcnt--;  /* Drop refcnt to 0; reincremented when I exit dheckxOpen. */
    }

    xTrace1(dheckxp,TR_EVENTS,"dheckxOpen returning session %x", dheckxSessn);
    return (dheckxSessn);
}


static XObj
dheckxCreateSessn( self, hlpRcv, hlpType, lls)
    XObj	self, hlpRcv, hlpType, lls;
{
    XObj	s;
    SState	*ss;
    PState	*ps = (PState *)self->state;
    Part        part[5];  /* So I can partInternalize partBuf, even if 5 participants. */
    char        partBuf[100];
    XObj        SessnArray[3];
    XObj        kms1;
    XObj        kms2;
    Part	part1[2];
    Part	part2[1];
    int		tmp;

    xTrace4(dheckxp, TR_EVENTS, "dheckxCreateSessn entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, lls);

    if (xControl(lls, GETPARTICIPANTS, partBuf, 100) <= 0)
    { xTrace0(dheckxp, TR_ERRORS, "dheckxCreateSessn: could not get participants");
      return ERR_XOBJ; }
	
    if ((tmp = partExtLen(partBuf)) > 5)
    { xTrace1(dheckxp, TR_ERRORS,
	      "dheckxCreateSessn: Too many %d participants; max 5.", tmp);
      return ERR_XOBJ; }

    partInternalize(part,partBuf);

    if (partLen(part) < 2)
    { xTrace0(dheckxp, TR_ERRORS, "dheckxCreateSessn: too few participants");
      return ERR_XOBJ; }

    if (!(ps->flags & ALL_PARTS)) /* open two key manager sessions */
    { partInit(part1,1);
      part1[0] = part[REMOTE_PART];
      part1[0].len = 1; /* major hack */
      kms1 = xOpen(self, self, xGetDown(self,KEY_MGR_PROTOCOL), part1);
      if (kms1 == ERR_XOBJ)
      { xTrace0(dheckxp, TR_ERRORS, "dheckxCreateSessn: could not open key manager");
	return ERR_XOBJ; }
      partInit(part2,1);
      part2[0] = part[LOCAL_PART];
      part2[0].len = 1; /* major hack */
      kms2 = xOpen(self, self, xGetDown(self,KEY_MGR_PROTOCOL), part2);
      if (kms2 == ERR_XOBJ)
      { xTrace0(dheckxp, TR_ERRORS, "dheckxCreateSessn: could not open key manager");
	xClose(kms1);
	return ERR_XOBJ; }
    }
    else    /* open one key manager session */
    { partInit(part1,2);
      part1[REMOTE_PART] = part[REMOTE_PART];
      part1[LOCAL_PART] = part[LOCAL_PART];
      part1[REMOTE_PART].len = part1[LOCAL_PART].len = 2;  /* ugly hack */
      kms1 = xOpen(self, self, xGetDown(self,KEY_MGR_PROTOCOL), part1);
      if (kms1 == ERR_XOBJ)
      { xTrace0(dheckxp, TR_ERRORS, "dheckxCreateSessn: could not open key manager");
	return ERR_XOBJ; }
    }

    SessnArray[0]=lls;
    SessnArray[1]=kms1;
    if (!(ps->flags & ALL_PARTS)) SessnArray[2]=kms2;
    s = xCreateSessn(dheckxSessnInit, hlpRcv, hlpType, self,
		     (ps->flags & ALL_PARTS) ? 2 : 3, SessnArray);
    s->binding = (Bind) mapBind(ps->actMap, (char *)&lls, (int)s);

    if ( s->binding == ERR_BIND ) {
	xTrace0(dheckxp, TR_ERRORS, "dheckxCreateSessn: mapBind error");
	xClose(kms1);
 	if (!(ps->flags & ALL_PARTS)) xClose(kms2);
	xDestroy(s);
	return ERR_XOBJ;
      }

    ss = X_NEW(SState);
    xTrace1(dheckxp, TR_FULL_TRACE, "dheckxCreateSessn: X_NEW(SState) returned %x", ss);
    bzero((void *)ss, sizeof(SState));
    ss->state = DHECKXNEW;
    s->state = (VOID *)ss;

    xTrace1(dheckxp, TR_MORE_EVENTS, "dheckxCreateSessn returning %x",s);
    return s;
}


static xkern_return_t
dheckxOpenEnable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(dheckxp, TR_EVENTS, "dheckxOpenEnable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    if ( defaultOpenEnable(ps->pasMap, hlpRcv, hlpType, 
	ps->ownhlptype?(VOID *)&self:(VOID *)&hlpType) == XK_FAILURE )
    { xTrace0(dheckxp, TR_MORE_EVENTS, "dheckxOpenEnable: defaultOpenEnable failed ");
      return XK_FAILURE; }
    
    val = xOpenEnable(self, ps->ownhlptype?self:hlpType, 
		      xGetDown(self,TRANSPORT_PROTOCOL), part);
    if ( val == XK_FAILURE )
    { xTrace0(dheckxp, TR_MORE_EVENTS, "dheckxOpenEnable: xOpenEnable failed ");
      return XK_FAILURE; }

    xTrace1(dheckxp, TR_MORE_EVENTS, "dheckxOpenEnable returning %x", val);
    return val;
} 


static xkern_return_t
dheckxOpenDisable(self, hlpRcv, hlpType, part)
    XObj self, hlpRcv, hlpType;
    Part *part;
{
    PState	*ps = (PState *)self->state;
    xkern_return_t val;

    xTrace4(dheckxp, TR_EVENTS, "dheckxOpenDisable entered: args %x %x %x %x",
	    self, hlpRcv, hlpType, part);

    val = defaultOpenDisable(ps->pasMap, hlpRcv, hlpType, 
			     ps->ownhlptype?(void*)&self:(void*)&hlpType);

    xTrace1(dheckxp, TR_MORE_EVENTS, "dheckxOpenDisable returning %x", val);
    return val;
}

static xkern_return_t
dheckxOpenDone(self, lls, llp, hlpType)
    XObj self, lls, llp, hlpType;
{
    XObj        s;
    PState      *pstate;
    Enable      *e;
   
    xTrace4(dheckxp, TR_MAJOR_EVENTS, "Entered dheckxOpenDone: args %x %x %x %x",
	    self, lls, llp, hlpType);

    if ( self == hlpType ) {
        xTrace0(dheckxp, TR_ERRORS, "self == hlpType in dheckxOpenDone");
      /*  return XK_FAILURE; */
    }
    pstate = (PState *)self->state;

    /*
     * check for open enables
     */
    if ( mapResolve(pstate->pasMap, &hlpType, &e) == XK_FAILURE ) {
        /*
         * This shouldn't happen
         */
        xTrace0(dheckxp, TR_ERRORS,
                "dheckxOpenDone: Couldn't find hlp for incoming session");
        return XK_FAILURE;
    }

    xDuplicate(lls);
    if ( (s = dheckxCreateSessn(self, e->hlpRcv, e->hlpType, lls))
                == ERR_XOBJ ) {
        xTrace0(dheckxp, TR_ERRORS,
                "dheckxOpenDone: Couldn't create session");
        xClose(lls);
        return XK_FAILURE;
    }

    if (((SState *)(s->state))->state == DHECKXNEW)
      ((SState *)(s->state))->state = DHECKXRESPONDER;
    else   /* Barf on any other state.  Some other path must have tried to use it? */
    { xTrace1(dheckxp, TR_ERRORS,
	      "dheckxOpenDone: Wrong state %d for newly created session.",
	      ((SState *)(s->state))->state);
      xClose(lls);
      return XK_FAILURE; }

    xTrace0(dheckxp, TR_EVENTS,
            "dheckxOpenDone: Passively opened session successfully created");
    return xOpenDone(e->hlpRcv, s, self);
}



static xkern_return_t
dheckxDemux( self, lls, msg )
    XObj	self, lls;
    Msg		*msg;
{
  PState	*ps = (PState *)self->state;
  XObj 	s = 0;
  xkern_return_t val;
  
  xTrace3(dheckxp, TR_EVENTS, "dheckxDemux entered: args %x %x %x",
	  self, lls, msg);

  xIfTrace(dheckxp, TR_DETAILED) msgShow(msg);

  if ( mapResolve(ps->actMap, &lls, &s) == XK_FAILURE ) { 
	xTrace0(dheckxp, TR_EVENTS, "dheckxDemux: map resolve fails");
  	return XK_FAILURE;
  }

  if ( xIsSession(s) )  val = xPop(s, lls, msg, 0);
  else val = XK_FAILURE;

  if (val == XK_SUCCESS)
  xTrace0(dheckxp, TR_FUNCTIONAL_TRACE, "dheckxDemux returning XK_SUCCESS");
  else if (val == XK_FAILURE)
  xTrace0(dheckxp, TR_FUNCTIONAL_TRACE, "dheckxDemux returning XK_FAILURE");
  else
  xTrace1(dheckxp, TR_FUNCTIONAL_TRACE, "dheckxDemux returning %x", val);

  return val;
}


static xkern_return_t
dheckxClose( s )
    XObj	s;
{
    PState	*ps = (PState *)xMyProtl(s)->state;
    xkern_return_t res;
    SState	*ss;

    xTrace1(dheckxp, TR_MAJOR_EVENTS, "dheckxClose: closing session %x", s);
    xAssert( xIsSession( s ) );
    xAssert( s->rcnt == 0 );
    res = mapRemoveBinding( ps->actMap, s->binding );
    xAssert( res != XK_FAILURE );
    /* Clear out SState. */
    if (ss = s->state)
    {
      /* I don't have to do anything about semaphore & event, because they
	 are only used while waiting for reply & computing, and nobody will
	 have the session pointer to call Close with. */
      bzero((void *)ss,sizeof(SState)); }   /* The xDestroy will xFree(ss). */
    xClose(xGetDown(s,TRANSPORT_PROTOCOL));
    xClose(xGetDown(s,KEY_MGR_SESS_OUT));  /* OUT==ALL */  /* Perhaps invalidate key before close? $$ */
    if (!(ps->flags & ALL_PARTS)) xClose(xGetDown(s,KEY_MGR_SESS_IN));  /* Perhaps invalidate key before close? $$ */
    xDestroy( s );
    xTrace0(dheckxp, TR_DETAILED, "dheckxClose returning");
    return XK_SUCCESS;
}

static
long dheckxcopy(to,from,len,ignore)
    void *to; char *from; void *ignore; long len;
{
  xTrace4(dheckxp, TR_FULL_TRACE, "dheckxcopy args: %x %x %x %x", to,from,len,ignore);

  bcopy(from,to,len);  return len;
}

static xmsg_handle_t
dheckxPush( s, msg )
    XObj	s;
    Msg 	*msg;
{ int val;

  xTrace2(dheckxp, TR_EVENTS, "dheckxPush: session %x  msg %x", s, msg);

  if (((SState *)(s->state))->state == DHECKXTRANSPARENT)
    val = xPush(xGetDown(s, TRANSPORT_PROTOCOL), msg);
  else { xTrace2(dheckxp, TR_ERRORS,
	"dheckxPush: Session isn't in transparent state %d; dropping message %x",
		 ((SState *)(s->state))->state, msg);
	 val = XK_FAILURE; }

  xTrace1(dheckxp, TR_DETAILED, "dheckxPush exiting, value %x", val);

  return val;
}


static xkern_return_t
dheckxPop( s, lls, m, hdr)
    XObj	s, lls;
    Msg 	*m;
    void        *hdr;
{ SState *ss = (SState *)s->state;

  xTrace5(dheckxp, TR_EVENTS, "dheckxPop: args %x %x %x %x.  state %d.",
	  s, lls, m, hdr, ss->state);
  switch (ss->state)
  { case DHECKXTRANSPARENT:  /* the ordinary case, after everything is set up */
      return xDemux(s, m);

    case DHECKXINITWAIT:     /* process the reply to our offer message */
      if (dheckxKeyExchangeDone(s, m) == XK_SUCCESS)
	{ dheckxOpenWakeup(0,s); return XK_SUCCESS; } /* if ok, wakeup Open */
      else return XK_FAILURE;
      /* Fail if something wrong with reply; Open will eventually time out anyway. */

    case DHECKXRESPONDER:    /* reply to an offer message */
      return dheckxKeyExchangeReply(s, m);
      /* Returns XK_FAILURE if offer has bad format. */

    /* Other possible states:
       REPLYING, if the replier is slow computing the key.
       COMPUTING, possibly. */

    default:
      xTrace2(dheckxp, TR_ERRORS,
	      "dheckxPop: Unexpected session state %x, dropping message %x",
	      ss->state, m);
      return XK_FAILURE;  }
}


static int
dheckxControlProtl( self, op, buf, len )
    XObj	self;
    int 	op, len;
    char 	*buf;
{ xTrace4(dheckxp, TR_EVENTS, "dheckxControlProtl: args %x %x %x %x", self, op, buf, len);
  xAssert(xIsProtocol(self));
  return xControl(xGetDown(self, TRANSPORT_PROTOCOL), op, buf, len);
}

static int
dheckxControlSessn(s, op, buf, len)
    XObj	s;
    int 	op, len;
    char 	*buf;
{ xTrace4(dheckxp, TR_EVENTS, "dheckxControlSessn: args %x %x %x %x", s, op, buf, len);
  xAssert(xIsSession(s));
  return xControl(xGetDown(s,TRANSPORT_PROTOCOL), op, buf, len);
}


static void
dheckxSessnInit(s)
    XObj s;
{ xTrace1(dheckxp, TR_EVENTS, "dheckxSessnInit: session %x", s);
  s->push = dheckxPush;
  s->close = dheckxClose;
  s->control = dheckxControlSessn;
  s->pop = dheckxPop;
}

