/*#include <sys/systeminfo.h>*/
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include "xkernel.h"
#include "ip.h"
#include "tcp.h"
#include "x_stdio.h"
#include "tcpp_internal.h"

#define msgsQueued( cs )    (cs->head != NULL)

typedef enum { 
  QPUMP_INACTIVE_S, QPUMP_OPENING_S, QPUMP_CLOSING_S, QPUMP_OK_S 
} ConnectStatus_s;

typedef enum { 
  QPUMP_INACTIVE=0, QPUMP_OPENING, QPUMP_FLUSHING, QPUMP_OK, QPUMP_QUEUED,
  QPUMP_NOT_OPENED
} ConnectStatus_r;

#define HOST_TYPE IPhost
#define QUEUE_MAX_SIZE 102400
#define BUFFER_SIZE 4096
#define MA_QUEUE_SIZE 4096
#define QUANTUM 1000
#define DELTA 20

#define TRACE_VAR qpumpp
#define TCP_NAME_S "tcp_s"
#define TCP_NAME_R "tcp_r"
#define MAP_SIZE 101
#define QUEUE_INTERVAL     2    /* delay for queueTimeout, in ms */
#define CLOSE_TICKS     2400    /* number of QUEUE_INTERVALs before close */

typedef struct {
  Protl tcp_s, tcp_r;
  Map   bindmap;
} PState;

typedef struct msgnode *MsgQ;

struct msgnode {
  MsgQ       next;
  Msg        *msg;            /* separate msg from node */
};

#define QPUMP_CLOSE_RCVD   1

typedef struct {
  ConnectStatus_s  status_s;
  ConnectStatus_r  status_r;
  u_char           flags;
  PState           *ps;
  IPhost           SourceAddr, DestAddr;
  long             SourcePort, DestPort;
  Sessn            lls, s_lls;
  struct msgnode   *head, *tail;
  int              queue_size;
  int              max_queue_size_reached;
  int              buffer_is_full;
  XTime            full_buffer_began;
  XTime            full_buffer_ended;
  long             full_buffer_total_time;
  long             average_packet_size;
  long             average_delay;
  long             number_of_packets;
  int              ma_timer_on;
  XTime            ma_delay_began;
  XTime            ma_delay_ended;
  long             ma_queue[MA_QUEUE_SIZE];     /* moving average queue */
  int              ma_next_element;
  XTime            quantum_start;
  int              cur_delay;  /* milliseconds */
  int              CloseCounter;
  Event            queueEvent;
} CState;

int traceqpumpp;

static xkern_return_t   qpumpDemux( Protl, Sessn, Msg * );
static xkern_return_t   openDone( Protl, Sessn, Protl, Protl );
static xkern_return_t   closeDone( Sessn );
static void             identifyDownProtocols ( Protl );
static void             handleOpenMsg ( Protl, Msg *, Sessn * );
static void             handleDataMsg ( Protl, Msg *, Sessn * );
static void             handleCloseMsg ( Protl, Msg *, Sessn * );
static void             QueueMsg ( CState *, Msg *);
static Msg              *DequeueMsg ( CState *);
void                    flushMsgQueue (CState *cs);
void                    CloseConnection (CState *cs);
void                    QueueTimeout (Event e, void *arg);
void                    StartMATimer(CState *cs);
void                    EndMATimer(CState *cs);
void                    DampingDelay(CState *cs);
void                    milli_sleep(long milliseconds);

int qpump_init (Protl);

int qpump_init (self)
     Protl self;
{
  PState	*ps;
  
  printf ("I am the Quantized Pump router, version 1.0 alpha\n");
  
  /* Initialize protocol method table */
  self->demux = qpumpDemux;
  self->opendone = openDone;
  self->closedone = closeDone;

  ps = X_NEW(PState);
  bzero((char *)ps, sizeof(PState));
  self->state = (VOID *)ps;

  identifyDownProtocols (self);
  
  ps->bindmap = mapCreate (MAP_SIZE, sizeof (u_int));

  return 0;
}


static void identifyDownProtocols (self)
     Protl self;
{
  Protl llp1, llp2, tcp;
  PState	*ps = (PState *)self->state;
  
  xTrace1(TRACE_VAR, TR_EVENTS, "Using \"%s\" for SENDER TCP name", TCP_NAME_S);
  tcp = xGetProtlByName (TCP_NAME_S);

  if (!xIsProtocol(tcp)) {
      Kabort ("Could not find a protocol object for SENDER TCP");
    }
  
  llp1 = xGetDown(self, 0);
  llp2 = xGetDown(self, 1);
  if (!xIsProtocol(llp1) || !xIsProtocol(llp2)) {
      Kabort("QPUMP does not have two lower protocols");
    }

  if (llp1 == tcp)
    {
      ps->tcp_s = llp1;
      ps->tcp_r = llp2;
    }
  else if (llp2 == tcp)
    {
      ps->tcp_r = llp1;
      ps->tcp_s = llp2;
    }
  else
    {
      Kabort("QPUMP is not configured over TCP");
    }
}



static xkern_return_t
openDone( self, s, llp, hlpType )
    Protl self, llp, hlpType;
    Sessn s;
{
  PState  *ps = (PState *)self->state;
  CState  *cs;
  u_short space = BUFFER_SIZE;
  Msg     m;
  IPpseudoHdr  IPhdr;
  int     i;
  
  xTrace0(TRACE_VAR, TR_MAJOR_EVENTS, "QPUMP openDone");

  if(ps->tcp_s != llp) {
    xTrace0(TRACE_VAR, TR_EVENTS, "Security violation attempted!");
    xTrace0(TRACE_VAR, TR_EVENTS,
	    "High security network trying to talk to the low security network.");
    xTrace0(TRACE_VAR, TR_EVENTS, "Access denied.");
    return XK_FAILURE;
  }
 
  if (mapResolve (ps->bindmap, &s, (void **)&cs) == XK_SUCCESS) {
      if (cs->status_s != QPUMP_OPENING_S)
        {
          xTrace0 (TRACE_VAR, TR_MAJOR_EVENTS,
                   "openDone: binding already exists");
          return XK_FAILURE;
        }
      return XK_SUCCESS;
    }
 
  xDuplicate(s);

  /* create appropriate state and enter into map */
  cs = X_NEW (CState);
  cs->status_s = QPUMP_OPENING_S;
  cs->s_lls = s;
  cs->head = cs->tail = NULL;
  cs->buffer_is_full = 0;
  cs->full_buffer_total_time = 0;
  cs->average_packet_size = 0;
  cs->average_delay = 0;
  cs->number_of_packets = 0;
  cs->queue_size = 0;
  cs->max_queue_size_reached = 0;
  xGetTime(&(cs->quantum_start));
  cs->cur_delay = 0;
  cs->status_r = QPUMP_OPENING;
  cs->ps = ps;
  cs->queueEvent = NULL;
  cs->CloseCounter = 0;

  for(i=0;i<MA_QUEUE_SIZE;i++)
    cs->ma_queue[i] = 0;
  cs->ma_next_element = 0;

  /* query lls to find out IP addresses */
  xControl (s, IP_GETPSEUDOHDR, (char *) &IPhdr, sizeof(IPpseudoHdr));

  cs->SourceAddr = IPhdr.dst;
  xControl (s, GETPEERPROTO, (char *) &(cs->SourcePort), sizeof(long));;
  cs->DestAddr = IPhdr.src;
  xControl (s, GETMYPROTO, (char *) &(cs->DestPort), sizeof(long));;
 
  if ( mapBind (ps->bindmap, (void *) &s, (void *) cs) == ERR_BIND)
    {
      xTrace0 (TRACE_VAR, TR_ERRORS, "openDone: binding failed");
      return XK_FAILURE;
    }
 
  /* set TCP buffer size */
  if (xControl(s, TCP_SETRCVBUFSIZE, (char*)&space, sizeof(space)) < 0)
    xError("saveServerSessn: TCP_SETRCVBUFSIZE failed");
  if (xControl(s, TCP_SETRCVBUFSPACE, (char*)&space, sizeof(space)) < 0)
      xError("saveServerSessn: TCP_SETRCVBUFSPACE failed");
 
  handleOpenMsg (self, &m, &s);
 
  cs->status_s = QPUMP_OK_S;
 
  return XK_SUCCESS;
}


static void 
handleOpenMsg ( self, dg, s )
     Protl self;
     Msg *dg;
     Sessn *s;
{
  PState	*ps = (PState *)self->state;
  CState *cs;
  Part  p[2];
  int   noblock = 1;

  xTrace0(TRACE_VAR, TR_MAJOR_EVENTS, "qpump received OpenMsg.");

  if (mapResolve(ps->bindmap, s, (void **)&cs) == XK_FAILURE) {
    printf("(!) Fatal error. Unknown connection!.\n");
    exit(-1);
    }

  /*
   * Initialize the connection state 
   */

  xTrace0 (TRACE_VAR, TR_EVENTS, "\nOpening connection");
  xTrace5 (TRACE_VAR, TR_EVENTS, 
	   "Source IP addr: %d.%d.%d.%d, TCP port: %d",
	   cs->SourceAddr.a, cs->SourceAddr.b, cs->SourceAddr.c, 
	   cs->SourceAddr.d, cs->SourcePort);
  xTrace5 (TRACE_VAR, TR_EVENTS, "Dest IP addr: %d.%d.%d.%d, TCP port: %d\n",
	   cs->DestAddr.a, cs->DestAddr.b, cs->DestAddr.c, 
	   cs->DestAddr.d, cs->DestPort);
  
  /* Open TCP connection with destination, save pointer to lls */
  partInit (p, 2);
  partPush (p[0], &(cs->DestAddr), sizeof(IPhost));
  partPush (p[0], &(cs->DestPort), sizeof(long));
  partPush (p[1], &(cs->SourceAddr), sizeof(IPhost));
  partPush (p[1], &(cs->SourcePort), sizeof(long));

  cs->lls = xOpen (self, self, ps->tcp_r, p);
  if ( cs->lls == ERR_SESSN ) 
    {
      xTrace0 (TRACE_VAR, TR_ERRORS, "Could not open R-side TCP connection");
      /* Take appropriate action here */
      cs->status_r = QPUMP_INACTIVE;
      printf ("ERROR: Unable to open R-side TCP connection!!!\n");
      CloseConnection (cs);
      return;
    }
  xTrace0(TRACE_VAR, TR_EVENTS, "R-side connection open");

  xControl (cs->lls, SETNONBLOCKINGIO, (char *) &noblock, sizeof(int));
  cs->queueEvent = evSchedule (QueueTimeout, cs, 10);

  /* 
   * flushMsgQueue() takes care of state changes, so call it even if
   * no messages are queued.
   */
  flushMsgQueue(cs);
}


xkern_return_t
qpumpDemux( self, lls, dg )
    Protl self;
    Sessn lls;
    Msg *dg;
{
  PState        *ps = (PState *)self->state;
  CState *cs;

  /* Retrieve state from map based on LLS */
  if (mapResolve(ps->bindmap, &lls, (void **)&cs) == XK_FAILURE) {
    xTrace0(TRACE_VAR, TR_EVENTS, "Security violation attempted!");
    xTrace0(TRACE_VAR, TR_EVENTS,
	    "High security network trying to talk to the low security network.");

    xTrace0(TRACE_VAR, TR_EVENTS, "Access denied.");
    return XK_FAILURE;
  }

  /* Increment ref cnt on lls */
  xDuplicate (lls);
 
  xTrace0(TRACE_VAR, TR_EVENTS, "Message received.");
  handleDataMsg (self, dg, &lls);
 
  return XK_SUCCESS;
}


static void 
handleDataMsg ( self, dg, s )
     Protl self;
     Msg *dg;
     Sessn *s;
{
  PState	*ps = (PState *)self->state;
  CState *cs;
  u_short len;

  if (mapResolve(ps->bindmap, s, (void **)&cs) == XK_FAILURE) {
    printf ("Error: (!) Got some data before the Open message!\n");
    return;
  }
  
  /*
   * Drop this message if the state is found but is inactive
   */
  if (cs->status_r == QPUMP_INACTIVE)
    return;

  /* reset CloseCounter (detects an idle connection) */
  cs->CloseCounter = 0;

  len = msgLen(dg);
  
  switch (cs->status_r)
    {
    case QPUMP_OPENING:
    case QPUMP_FLUSHING:
      /* queue message for later transmission */
      xTrace0 (TRACE_VAR, TR_EVENTS,  "Got message -> queued (flushing state).");
      QueueMsg (cs, dg);
      break;

    case QPUMP_QUEUED:
      /* 
       * Message to go out is at the front of the queue.
       * Queue this message, and then try to flush the queue.
       */
      xTrace0 (TRACE_VAR, TR_EVENTS,  "Got message -> queued (queued state).");
      QueueMsg (cs, dg);
      break;

    case QPUMP_OK:
      xTrace0 (TRACE_VAR, TR_EVENTS, "Got message -> queuing (empty queue).");
      QueueMsg (cs, dg);
      cs->status_r = QPUMP_QUEUED;
      break;

    default:
      ;
    }

  if(cs->queue_size < QUEUE_MAX_SIZE) {
    DampingDelay(cs);
    xControl(cs->s_lls, TCP_SETRCVBUFSPACE, (char *)&len, sizeof(len));
    xTrace1(TRACE_VAR, TR_EVENTS, "Receive window reopened by %d bytes", len);
  }
}


static xkern_return_t
closeDone( lls )
    Sessn        lls;
{
  PState        *ps;
  CState        *cs;
  Protl          self;
  Msg           m;
 
  xTrace0(TRACE_VAR, TR_MAJOR_EVENTS, "QPUMP closeDone called");
 
  /* Retrieve state from map based on LLS */
  self = xGetUp (lls);
  ps = (PState *)self->state;
  if (mapResolve(ps->bindmap, &lls, (void **)&cs) == XK_FAILURE) {
    xTrace0(TRACE_VAR, TR_EVENTS, "Security violation attempted!");
    xTrace0(TRACE_VAR, TR_EVENTS,
	    "High security network trying to talk to the low security network.");
    xTrace0(TRACE_VAR, TR_EVENTS, "Access denied.");
    return XK_FAILURE;
  }
 
  if (cs->status_s == QPUMP_CLOSING_S)
    return XK_SUCCESS;
 
  cs->status_s = QPUMP_CLOSING_S;

  printf("Max queue size reached: %d\n", cs->max_queue_size_reached);
  printf("Total time spent with buffer full: %d\n", cs->full_buffer_total_time);
  printf("Average packet size: %d\n",
	 cs->average_packet_size/cs->number_of_packets);
  printf("Average delay: %d\n", cs->average_delay/cs->number_of_packets);

  handleCloseMsg (self, &m, &lls);
 
  return XK_SUCCESS;
}


static void 
handleCloseMsg ( self, dg, s )
     Protl self;
     Msg *dg;
     Sessn *s;
{
  PState	*ps = (PState *)self->state;
  CState *cs;

  if (mapResolve(ps->bindmap, s, (void **)&cs) == XK_FAILURE) {
      xTrace0 (TRACE_VAR, TR_ERRORS, 
	       "received Close message belongs to no known connection");
      return;
    }
  
  /* 
   * Remember that we've seen a close message for this connection, but 
   * don't close until all sequence numbers before this msg have been
   * sent.
   */
  if (cs->status_r != QPUMP_INACTIVE)
      cs->flags |= QPUMP_CLOSE_RCVD;

  /* Try to flush the queue and close now */
  if (cs->status_r == QPUMP_OK)
    CloseConnection(cs);
}


void flushMsgQueue (CState *cs)
{
  Msg *m;
  xkern_return_t status;
  int buffer_full;
  u_short len;

  xTrace0(TRACE_VAR, TR_EVENTS, "Flushing queue...");
  
  if (cs->status_r == QPUMP_NOT_OPENED)
    return;
  cs->status_r = QPUMP_FLUSHING;
  
  /*  while (msgsQueued(cs)) {*/
    if(cs->queue_size >= QUEUE_MAX_SIZE)
      buffer_full = 1;
    else
      buffer_full = 0;
    m = DequeueMsg (cs);
    len = msgLen(m);
    StartMATimer(cs);
    status = xPush (cs->lls, m);
    if (status == XMSG_ERR_WOULDBLOCK)
      {
	xTrace0 (TRACE_VAR, TR_EVENTS, "Message would block -- requeued");
	QueueMsg (cs, m);
	/*	break;*/

	cs->status_r = QPUMP_QUEUED;
	return;
      }
    EndMATimer(cs);
    xTrace0 (TRACE_VAR, TR_EVENTS, "Sent message from the queue");
    if(buffer_full) {
      DampingDelay(cs);
      xControl(cs->s_lls, TCP_SETRCVBUFSPACE, (char *)&len, sizeof(len));
      xTrace1(TRACE_VAR, TR_EVENTS, "Receive window reopened by %d bytes", len);
    }
    msgDestroy (m);
    /*  }*/

  if (msgsQueued(cs)) {
    xTrace0(TRACE_VAR, TR_EVENTS, "Done flushing queue -- some messages still left on the queue.");
    cs->status_r = QPUMP_QUEUED;
  } else {
    xTrace0(TRACE_VAR, TR_EVENTS, "Done flushing queue -- queue empty.");
    if(cs->flags & QPUMP_CLOSE_RCVD) CloseConnection(cs);
    cs->status_r = QPUMP_OK;
  }
}


void CloseConnection (CState *cs)
{
  Msg *m;

  if (cs->queueEvent != NULL)
    {
      evCancel (cs->queueEvent);
      cs->queueEvent = NULL;
      xTrace0(TRACE_VAR, TR_EVENTS, "Queue event canceled");
    }
  
  m = DequeueMsg (cs);
  while (m != NULL)
    {
      msgDestroy (m);
      m = DequeueMsg (cs);
    }

  /* Remove state from map */
  if ( mapUnbind (cs->ps->bindmap, &(cs->s_lls)) == XK_FAILURE )
    {
      xTrace0 (TRACE_VAR, TR_ERRORS, "mapUnbind failed");
    }
  
  if (cs->status_r != QPUMP_INACTIVE)
    xClose (cs->lls);
 
  free (cs);
}


static void QueueMsg ( cs, m )
     CState *cs;
     Msg    *m;
{
  MsgQ   n = X_NEW (struct msgnode);
  Msg    *new = X_NEW(Msg);
  int    len;

  len = msgLen(m);

  xTrace0(TRACE_VAR, TR_EVENTS, "Enqueueing message...");
  msgConstructCopy (new, m);
  
  if ( cs->head == NULL ) {
    cs->head = n;
    cs->tail = n;
  }
  else {
    cs->tail->next = n;
    cs->tail = n;
  }

  n->msg = new;
  n->next = NULL;

  cs->queue_size += len;
  if(cs->queue_size > cs->max_queue_size_reached)
    cs->max_queue_size_reached = cs->queue_size;

  if((cs->queue_size > QUEUE_MAX_SIZE) && (!cs->buffer_is_full)) {
    xGetTime(&(cs->full_buffer_began));
    cs->buffer_is_full = 1;    
  }

  cs->number_of_packets++;
  cs->average_packet_size += len;
}


static Msg *DequeueMsg ( cs )
     CState *cs;
{
  MsgQ   n;
  Msg    *m;
  int    len;
  XTime time_diff;

  if (cs->head != NULL) {
    xTrace0(TRACE_VAR, TR_EVENTS, "Dequeueing message...");
    n = cs->head;
    cs->head = cs->head->next;
    m = n->msg;
    free (n);
    if(cs->head==NULL)
      cs->tail = NULL;
    len = msgLen(m);
    cs->queue_size -= len;
    return m;
  }

  if((cs->queue_size < QUEUE_MAX_SIZE) && (cs->buffer_is_full)) {
    xGetTime(&(cs->full_buffer_ended));
    cs->buffer_is_full = 0;

    xSubTime(&time_diff, cs->full_buffer_ended, cs->full_buffer_began);

    cs->full_buffer_total_time += (time_diff.sec)*1000 + (time_diff.usec)/1000; 
  }

  return NULL;
}


void QueueTimeout (Event e, void *arg)
{
  CState *cs = (CState *) arg;
  
  if (evIsCancelled (e))
    return;

  while (cs->queueEvent != NULL)
    {
      xTrace0 (TRACE_VAR, TR_EVENTS, "Queue timeout...");

      if (!(cs->flags & QPUMP_CLOSE_RCVD))
	{
	  if (++cs->CloseCounter >= CLOSE_TICKS)
	    {
	      xTrace0 (TRACE_VAR, TR_MAJOR_EVENTS,
		       "Connection is idle -- closing");
	      cs->flags |= QPUMP_CLOSE_RCVD;
	      CloseConnection (cs);
	      evDetach(e);
	      return;
	    }
	}
  
      if (cs->status_r == QPUMP_QUEUED) {
	flushMsgQueue (cs);
	flushMsgQueue (cs);
      } else {
	xTrace1 (TRACE_VAR, TR_EVENTS, "State is %d", cs->status_r);
      }

      Delay(QUEUE_INTERVAL);
    }
  
  evDetach (e);
}

void StartMATimer(CState* cs) {

  if(!(cs->ma_timer_on)) {
    xGetTime(&(cs->ma_delay_began));
    cs->ma_timer_on = 1;
  }
}

void EndMATimer(CState* cs) {
  XTime time_diff;
  long msec_delay;      /* Milliseconds */

  xGetTime(&(cs->ma_delay_ended));
  cs->ma_timer_on = 0;

  xSubTime(&time_diff, cs->ma_delay_ended, cs->ma_delay_began);

  msec_delay = (time_diff.sec)*1000 + (time_diff.usec)/1000;

  cs->ma_queue[cs->ma_next_element] = msec_delay;
  cs->ma_next_element++;

  cs->average_delay += msec_delay;
}

void DampingDelay(CState* cs) {
  XTime quantum_end, quantum_diff;
  long quantum_diff_milliseconds;
  long average_delay;   /* Milliseconds */
  int i;

  xGetTime(&quantum_end);
  xSubTime(&quantum_diff, quantum_end, cs->quantum_start);

  quantum_diff_milliseconds = (quantum_diff.sec*1000)+(quantum_diff.usec/1000);

  if(quantum_diff_milliseconds >= QUANTUM) {
    average_delay = 0;
    for(i=0;i<cs->ma_next_element;i++)
      average_delay += cs->ma_queue[i];
    if(average_delay)
      average_delay = average_delay / cs->ma_next_element;

    if(average_delay >= cs->cur_delay)
      cs->cur_delay += DELTA;
    else
      cs->cur_delay -= DELTA;

    if(cs->cur_delay < 0)
      cs->cur_delay = 0;

    cs->ma_next_element = 0;
    xGetTime(&(cs->quantum_start));
  }

  milli_sleep(cs->cur_delay);

  /* printf("%d\n", cs->queue_size);*/
}

void milli_sleep(long milliseconds) {
  XTime wait_start, wait_end, wait_diff;
  long cur_milliseconds;

  xGetTime(&wait_start);

  for(;;) {
    xGetTime(&wait_end);
    xSubTime(&wait_diff, wait_end, wait_start);
    cur_milliseconds = (wait_diff.sec*1000)+(wait_diff.usec/1000);
    if(cur_milliseconds >= milliseconds)
      return;
  }
}
