/*
 * tcp_subr.c
 *
 * Derived from:
 *
 * Copyright (c) 1982, 1986 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of California at Berkeley. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 *
 *	@(#)tcp_subr.c	7.13 (Berkeley) 12/7/87
 *
 * Modified for x-kernel v3.2
 * Modifications Copyright (c) 1993,1991,1990  Arizona Board of Regents
 *
 * $Revision: 1.1 $
 * $Date: 1996/07/11 21:40:04 $
 */

#include "xkernel.h"
#include "ip.h"
#include "icmp.h"
#include "tcp_internal.h"
#include "tcp_fsm.h"
#include "tcp_seq.h"
#include "tcp_timer.h"
#include "tcp_var.h"
#include "tcpip.h"

int	tcp_r_ttl = TCP_R_TTL;

/*
 * Create template to be used to send tcp_r packets on a connection.
 * Call after host entry created, allocates an mbuf and fills
 * in a skeletal tcp_r/ip header, minimizing the amount of work
 * necessary when the connection is used.
 */
struct tcp_riphdr *
tcp_r_template(tp)
	struct tcp_rcb *tp;
{
	register struct inpcb *inp = tp->t_inpcb;
	register struct tcp_riphdr *n;
	XObj	lls, tcp_rSessn;
	IPpseudoHdr	pHdr;

	if ((n = tp->t_template) == 0) {
		n = (struct tcp_riphdr *)xMalloc(sizeof *n);
	}
	tcp_rSessn = tcp_rcbtoso(tp);
	xAssert(xIsSession(tcp_rSessn));
	lls = xGetDown(tcp_rSessn,0);
	xAssert(xIsSession(lls));
	if ( xControl(lls, IP_GETPSEUDOHDR, (char *)&pHdr, sizeof(pHdr)) < 0 ) {
	    xTraceP0(tcp_rSessn, TR_ERRORS,
		     "tcp_r_template could not get pseudo-hdr from lls");
	    pHdr.prot = 0;
	}
	n->ti_next = n->ti_prev = 0;
	n->ti_x1 = 0;
	n->ti_pr = pHdr.prot;
	n->ti_len = sizeof (struct tcp_riphdr) - sizeof (struct ipovly);
	n->ti_src = inp->inp_laddr;
	n->ti_dst = inp->inp_raddr;
	n->ti_sport = inp->inp_lport;
	n->ti_dport = inp->inp_rport;
	n->ti_seq = 0;
	n->ti_ack = 0;
	n->ti_x2 = 0;
	n->ti_off = 5;
	n->ti_flags = 0;
	n->ti_win = 0;
	n->ti_sum = 0;
	n->ti_urp = 0;
	/*
	 * IP pseudo-header
	 */
	n->ti_p.src = *(IPhost *)&n->ti_src;
	n->ti_p.dst = *(IPhost *)&n->ti_dst;
	n->ti_p.zero = 0;
	n->ti_p.prot = n->ti_pr;
	return (n);
}

/*
 * Send a single message to the TCP_R at address specified by
 * the given TCP_R/IP header.  If flags==0, then we make a copy
 * of the tcp_riphdr at ti and send directly to the addressed host.
 * This is used to force keep alive messages out using the TCP_R
 * template for a connection tp->t_template.  If flags are given
 * then we send a message back to the TCP_R which originated the
 * segment ti, and discard the mbuf containing it and any other
 * attached mbufs.
 *
 * In any case the ack and sequence number of the transmitted
 * segment are as specified by the parameters.
 */
void
tcp_r_respond(tp, th, pHdr, ack, seq, flags, tcp_r)
	struct tcp_rcb *tp;
    	register struct tcp_rhdr *th;
    	register IPpseudoHdr *pHdr;
	tcp_r_seq ack, seq;
	int flags;
	XObj tcp_r;
{
	int win = 0, tlen;
	Msg m;
	XObj s = NULL;
	struct tcp_rhdr tHdr;
	hdrStore_t	store;

	if (tp) {
		s = tp->t_inpcb->inp_session;
		win = sototcp_rst(s)->rcv_space;
		s = xGetDown(s, 0);
	}
#ifdef TCP_R_COMPAT_42
	tlen = flags == 0;
#else
	tlen = 0;
#endif
	tHdr = *th;
	if (flags == 0) {
		flags = TH_ACK;
	} else {
	    	u_short tmp;

		tmp = tHdr.th_sport;
		tHdr.th_sport = tHdr.th_dport;
		tHdr.th_dport = tmp;
	}
	tHdr.th_seq = seq;
	tHdr.th_ack = ack;
	tHdr.th_x2 = 0;
	tHdr.th_off = sizeof (struct tcp_rhdr) >> 2;
	tHdr.th_flags = flags;
	tHdr.th_win = win;
	tHdr.th_urp = 0;
	msgConstructEmpty(&m);
	
	store.m = &m;
	store.h = pHdr;
	msgPush(&m, tcp_rHdrStore, &tHdr, sizeof(tHdr), &store);
	if (s) {
	    xPush(s, &m);
	} else {
	    Part p[2];
	    
	    partInit(p, 2);
	    partPush(p[0], &pHdr->dst, sizeof(IPhost));
	    partPush(p[1], &pHdr->src, sizeof(IPhost));
	    s = xOpen(tcp_r, tcp_r, xGetDown(tcp_r, 0), p);
	    if ( xIsSession(s) ) {
		xPush(s, &m);
		xClose(s);
	    } else {
		xError("tcp_r_respond could not open lower session!");
	    }
	}
	msgDestroy(&m);
}


/*
 * Create a new TCP_R control block, making an
 * empty reassembly queue and hooking it to the argument
 * protocol control block.  Assumes that the ip control block and associated
 * tcp_r session already exist.
 */
struct tcp_rcb *
tcp_r_newtcp_rcb(inp)
	struct inpcb *inp;
{
	register struct tcp_rcb *tp;

	tp = (struct tcp_rcb *)xMalloc(sizeof *tp);
	bzero((char*)tp, sizeof *tp);
	tp->seg_next = tp->seg_prev = (struct reass *)tp;
	tp->t_maxseg = TCP_R_MSS;
	tp->t_flags = 0;		/* sends options! */
	tp->t_inpcb = inp;
	/*
	 * Init srtt to TCP_RTV_SRTTBASE (0), so we can tell that we have no
	 * rtt estimate.  Set rttvar so that srtt + 2 * rttvar gives
	 * reasonable initial retransmit time.
	 */
	tp->t_srtt = TCP_RTV_SRTTBASE;
	tp->t_rttvar = TCP_RTV_SRTTDFLT << 2;
	TCP_RT_RANGESET(tp->t_rxtcur, 
	    ((TCP_RTV_SRTTBASE >> 2) + (TCP_RTV_SRTTDFLT << 2)) >> 1,
	    TCP_RTV_MIN, TCP_RTV_REXMTMAX);
	tp->snd_cwnd = sbspace(sototcp_rst(inp->inp_session)->snd);
	tp->snd_ssthresh = 65535;		/* XXX */
	inp->inp_ppcb = (caddr_t)tp;
	return (tp);
}

/*
 * Drop a TCP_R connection, reporting
 * the specified error.  If connection is synchronized,
 * then send a RST to peer.
 */
struct tcp_rcb *
tcp_r_drop(tp, errnum)
	register struct tcp_rcb *tp;
	int errnum;
{
	XObj so = tp->t_inpcb->inp_session;
	PSTATE	*ps = (PSTATE *)xMyProtl(so)->state;

	xTrace2(tcp_rp, 2, "tcp_r_drop: tcp_rcb %X s %X", tp, so);
	if (TCP_RS_HAVERCVDSYN(tp->t_state)) {
	  tp->t_state = TCP_RS_CLOSED;
	  (void) tcp_r_output(tp);
	  ps->tcp_rstat->tcp_rs_drops++;
	} else {
	  ps->tcp_rstat->tcp_rs_conndrops++;
	}
	socantrcvmore_(so);
#if 0
	return (tcp_r_destroy(tp));
#else
	return 0;
#endif
}

/*
 * Close a TCP_R control block:
 *	discard all space held by the tcp_r
 *	discard internet protocol block
 *	wake up any sleepers
 */
struct tcp_rcb *
tcp_r_destroy(tp)
	register struct tcp_rcb *tp;
{
	register struct reass *this, *next;
	struct inpcb *inp;
	XObj so;
	PSTATE *pstate;

	xTrace1(tcp_rp, 1, "tcp_r_destroy: tcp_rcb %X", tp);
	inp = tp->t_inpcb;
	xTrace1(tcp_rp, 3, "tcp_r_destroy: inpcb %X", inp);
	so = inp->inp_session;
	pstate = (PSTATE *)so->myprotl->state;
	this = tp->seg_next;
	while (this != (struct reass *)tp) {
		next = this->next;
		msgDestroy(&this->m);
		remque(this);
		xFree((char *)this);
		this = next;
	}
	if (tp->t_template) {
	    tcp_rReleasePort(pstate->portstate, tp->t_template->ti_sport);
	    xFree((char *)tp->t_template);
	}
	xFree((char *)tp);
	inp->inp_ppcb = 0;
	/*
	 * This used to be a soisdisconnected, but really needs to delete
	 * all of the state.  
	 */
	delete_tcp_rstate(sototcp_rst(so));
	so->state = 0;
	mapRemoveBinding(pstate->activeMap, so->binding);
	{
		int i;

		for (i=0; i < so->numdown; i++) {
			xClose(xGetDown(so, i));
		}
	}
	xDestroy(so);

	in_pcbdetach_(inp);
	pstate->tcp_rstat->tcp_rs_closed++;
	return ((struct tcp_rcb *)0);
}


/*
 * Notify a tcp_r user of an asynchronous error;
 * just wake up so that he can collect error status.
 */
void
tcp_r_notify(inp)
	register struct inpcb *inp;
{

	sorwakeup_(inp->inp_session);
	sowwakeup_(inp->inp_session);
}


#if BSD<43
/* XXX fake routine */
tcp_r_abort(inp)
	struct inpcb *inp;
{
	return;
}
#endif

/*
 * When a source quench is received, close congestion window
 * to one segment.  We will gradually open it again as we proceed.
 */
void
tcp_r_quench(inp)
	struct inpcb *inp;
{
	struct tcp_rcb *tp = intotcp_rcb(inp);

	if (tp)
		tp->snd_cwnd = tp->t_maxseg;
}
